@@ -13,14 +13,14 @@ check_default_id()
 {
 	id=$(run_check_stdout $SUDO_HELPER "$TOP/btrfs" subvolume get-default .) \
 		|| { echo "$id"; exit 1; }
-	if $(echo "$id" | grep -vq "ID $1"); then
+	if echo "$id" | grep -vq "ID $1"; then
 		_fail "subvolume get-default: default id is not $1, but $id"
 	fi
 }
 
 run_check_mkfs_test_dev
 run_check_mount_test_dev
-cd "$TEST_MNT"
+cd "$TEST_MNT" || _fail "Cannot cd into TEST_MNT $TEST_MNT"
 
 check_default_id 5
 
